# -*- coding:utf-8 -*-

import six
import re
from datetime import datetime, timedelta, date
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse
import numpy as np
from .common import return_as_type

__all__ = ['parse_datetime', 'parse_timedelta', 'datetime_range', 'T', 'TD', 'TR', 'datetime', 'timedelta', 'date', 'relativedelta']


def parse_datetime(timestr, force_datetime=True):
    """Try parse timestr or integer or list of str/integer into datetimes
    timestr: single value or seq of:
                int/str: YYYYMMDD[HH[MM[SS]]] , delimiters are also allowed,
                    e.g.: YYYY-MM-DD HH:MM:SS.
                    Julian Days, e.g., YYYYJJJ[...] is also OK.
                datetime/date
    force_datetime:
                if True: return datetime even if input is date.
                if False: return date if input is date.
    """
    numonly_pattern = re.compile(r'\D*')
    jfmtstr_all = '%Y%j%H%M%S'
    gfmtstr_all = '%Y%m%d%H%M%S'

    if isinstance(timestr, (list, tuple)):
        timestrs = timestr
        return_type = 'list'
    elif isinstance(timestr, np.ndarray):
        timestrs = timestr
        return_type = 'np_arr'
    else:
        timestrs = [timestr]
        return_type = 'scalar'

    res_list = []
    for timestr in timestrs:
        if isinstance(timestr, (datetime, )):
            res_list.append(timestr)
            continue
        if isinstance(timestr, (date, )):
            if force_datetime == True:
                res_list.append(datetime.fromordinal(timestr.toordinal()))
            else:
                res_list.append(timestr)
            continue
        if isinstance(timestr, six.integer_types + (np.integer, )):
            timestr = str(timestr)

        # handle year or year-month format like YYYY, YYYY-mm etc
        if re.match(r'^\d{4}$', timestr):  # year
            timestr = timestr + '0101'
        else:
            month_m = re.match(r'^(\d{4})([\t\n\r\f-/]*)(\d{2})$', timestr)
            if month_m:
                timestr = '{}{}01'.format(month_m.group(1), month_m.group(3))

        if len(timestr) == 10:
            if re.match(r'^\d{10}$', timestr):
                timestr = timestr + '00'
            elif re.match(r'^\d{4}.\d{2}.\d{2}$', timestr):
                timestr = timestr + ' 00:00:00'

        try:
            res = parse(timestr)
        except ValueError:
            try:
                numonly_str = re.sub(numonly_pattern, '', timestr)
                nl = len(numonly_str)
                if nl >= 7:
                    if nl % 2 == 1:   #YYYYJJJ[...]
                        fmtstr = jfmtstr_all[:4+nl-7]
                    else:             #YYYYMMDD[...]
                        fmtstr = gfmtstr_all[:6+nl-8]
                    res = datetime.strptime(numonly_str, fmtstr)
                else:
                    res = parse(numonly_str)
            except Exception:
                res = None
        res_list.append(res)

    return return_as_type(res_list, return_type)


_tdelta_dict = {
    'd': 'days', 'h': 'hours', 'm': 'minutes', 's': 'seconds',  # legacy formats
    # pandas freq compatible
    'T': 'minutes', 'min': 'minutes',
    'H': 'hours', 'D': 'days', 'S': 'seconds',
}
_rel_tdelta_dict = {'Y': 'years', 'M': 'months', 'W': 'weeks'}
def parse_timedelta(timestr):
    """Try parse single value or seq of timestr/integer/timedelta into timedeltas.
    e.g. parse_timedelta(['30m', '2h', '1d', '15s', '2:30', '1:12:25']).
    Y: years
    M: months
    W: weeks
    d, D: days
    h, H: hours
    m, min, T: minutes
    s, S: seconds (default)
    """
    if isinstance(timestr, (list, tuple)):
        timestrs = timestr
        return_type = 'list'
    elif isinstance(timestr, np.ndarray):
        timestrs = timestr
        return_type = 'np_arr'
    else:
        timestrs = [timestr]
        return_type = 'scalar'

    res_list = []
    for timestr in timestrs:
        if timestr is None:
            res_list.append(None)
            continue
        if isinstance(timestr, (timedelta, relativedelta)):
            res_list.append(timestr)
            continue
        if isinstance(timestr, six.integer_types + (np.integer, )):
            timestr = str(timestr)
        timestr = timestr.strip()

        if ':' in timestr:
            tokens = timestr.split(':')
            seconds = 0
            for i, tk in enumerate(reversed(tokens)):
                seconds += int(tk) * (60 ** i)
            res = timedelta(seconds=seconds)
            res_list.append(res)
            continue

        m = re.match(r'([-+0-9]*)([A-Za-z]*)', timestr)
        if not m:
            res = None
        else:
            if m.group(1) == '':
                value = 1
            else:
                value = int(m.group(1))
            symbol = m.group(2)
            if symbol in _rel_tdelta_dict:
                res = relativedelta(**{_rel_tdelta_dict[symbol]: value})
            else:
                if symbol == '':
                    symbol = 's'
                key = _tdelta_dict.get(symbol)
                if key is None:
                    res = None
                else:
                    res = timedelta(**{key: value})
        res_list.append(res)

    return return_as_type(res_list, return_type)


def datetime_range(beg, end=None, tdelta='1h'):
    """Returns a list of datetimes from beg to end with tdelta.
    """
    if end is None:
        tokens = beg.split(':')
        beg = tokens[0]
        end = tokens[1]
        if len(tokens) >= 3:  # beg:end:td
            tdelta = tokens[2]

    if not isinstance(beg, datetime):
        beg = parse_datetime(beg)
    if not isinstance(end, datetime):
        end = parse_datetime(end)
    if not isinstance(tdelta, (timedelta, relativedelta)):
        tdelta = parse_timedelta(tdelta)
    result = []
    now = beg
    while now < end:
        result.append(now)
        now = now + tdelta
    return result


T = parse_datetime
TD = parse_timedelta
TR = datetime_range



# -*- coding:utf-8 -*-

import six
import unittest
from hypothesis import given
import hypothesis.strategies as st
import hypothesis.extra.numpy as np_st

import numpy as np
from dt_utils.parsers import *


class TestParseTDelta(unittest.TestCase):
    def test_legacy_plain_format(self):
        self.assertEqual(TD('15m'), timedelta(minutes=15))
        self.assertEqual(TD('1d'), timedelta(days=1))
        self.assertEqual(TD('2h'), timedelta(hours=2))
        self.assertEqual(TD('30s'), timedelta(seconds=30))
        self.assertEqual(TD('60'), timedelta(seconds=60))

    def test_legacy_relative_format(self):
        self.assertEqual(TD('2M'), relativedelta(months=2))
        self.assertEqual(TD('1Y'), relativedelta(years=1))
        self.assertEqual(TD('7W'), relativedelta(weeks=7))

    def test_pd_compatible_format(self):
        self.assertEqual(TD('15T'), timedelta(minutes=15))
        self.assertEqual(TD('15min'), timedelta(minutes=15))
        self.assertEqual(TD('1D'), timedelta(days=1))
        self.assertEqual(TD('2H'), timedelta(hours=2))
        self.assertEqual(TD('30S'), timedelta(seconds=30))

    def test_without_number(self):
        self.assertEqual(TD('T'), timedelta(minutes=1))
        self.assertEqual(TD('min'), timedelta(minutes=1))
        self.assertEqual(TD('D'), timedelta(days=1))
        self.assertEqual(TD('H'), timedelta(hours=1))
        self.assertEqual(TD('S'), timedelta(seconds=1))
        self.assertEqual(TD('m'), timedelta(minutes=1))
        self.assertEqual(TD('d'), timedelta(days=1))
        self.assertEqual(TD('h'), timedelta(hours=1))
        self.assertEqual(TD('s'), timedelta(seconds=1))
        self.assertEqual(TD(''), timedelta(seconds=1))

    def test_number(self):
        self.assertEqual(TD(3600), timedelta(hours=1))

    def test_hh_mm_ss(self):
        self.assertEqual(TD('1:20:35'), timedelta(hours=1) + timedelta(minutes=20) + timedelta(seconds=35))
        self.assertEqual(TD('1:20'), timedelta(minutes=1) + timedelta(seconds=20))

    def test_invalid(self):
        self.assertIsNone(TD('1what'))
        self.assertIsNone(TD('ablaksjedoij1i2jalskjdf'))
        self.assertIsNone(TD(None))

    def test_list(self):
        self.assertListEqual(TD([
            '15m', '1H', 3600, None
        ]), [
            timedelta(minutes=15), timedelta(hours=1), timedelta(seconds=3600), None
        ])

    def test_array(self):
        result = TD(np.array([
            '15m', '1H', 3600, None
        ]))
        self.assertIsInstance(result, np.ndarray)
        self.assertListEqual(list(result), [
            timedelta(minutes=15), timedelta(hours=1), timedelta(seconds=3600), None
        ])


if __name__ == '__main__':
    unittest.main()


import unittest
from dt_utils.parsers import *
from dt_utils.neat import *


class TestNeat(unittest.TestCase):
    def test_latest_neat(self):
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='T'), T(202405010947))
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='10T'), T(202405010940))
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='15T'), T(202405010945))
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='1h'), T(2024050109))
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='1D'), T(20240501))
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='M'), T(20240501))
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='1M'), T(20240501))
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='3M'), T(20240401))
        self.assertEqual(get_latest_neat_time(T(202405010947), freq='Y'), T(20240101))
        self.assertEqual(get_latest_neat_time(T(202305010947), freq='4Y'), T(20200101))

    def test_nearest_neat(self):
        self.assertEqual(get_nearest_neat_time(T(202405010947), freq='15T'), T(202405010945))
        self.assertEqual(get_nearest_neat_time(T(202405010947), freq='10T'), T(202405010950))
        self.assertEqual(get_nearest_neat_time(T(202405010947), freq='1M'), T(20240501))
        self.assertEqual(get_nearest_neat_time(T(202405210947), freq='1M'), T(20240601))
        self.assertEqual(get_nearest_neat_time(T(202105210947), freq='4Y'), T(20200101))
        self.assertEqual(get_nearest_neat_time(T(202305210947), freq='4Y'), T(20240101))

    def test_neat_beg_end(self):
        # TODO: test
        pass


if __name__ == '__main__':
    unittest.main()



__all__ = ['return_as_type']

import numpy as np


def return_as_type(seq, return_type=None):
    if return_type == 'scalar':
        return seq[0]
    elif return_type == 'np_arr':
        return np.array(seq)
    else:
        return seq



# -*- coding:utf-8 -*-

from datetime import datetime, timedelta
from time import sleep

__all__ = ['RetryTimer']


class RetryTimer(object):
    def __init__(self, max_wait=None, interval=None):
        self.max_wait = max_wait
        self.interval = interval

        self.next_interval = 0.0
        self.start_time = None

    def retry(self):
        if self.start_time is None:
            self.start_time = datetime.now()

        sleep(self.next_interval)
        if self.interval is not None:
            self.next_interval = self.interval
        return self.check()

    def check(self):
        if self.max_wait is None:
            return True

        if self.start_time is None:
            return True
        elif self.start_time + timedelta(seconds=self.max_wait) >= datetime.now():
            return True
        else:
            return False



import unittest
from dt_utils.parsers import T
from dt_utils.format import *

import numpy as np


class TestFormat(unittest.TestCase):
    def test_minimal_timestr(self):
        self.assertEqual(minimal_timestr(T(202305010000)), '20230501')
        self.assertEqual(minimal_timestr(T(202305010800)), '2023050108')
        self.assertEqual(minimal_timestr(T(20240601231500)), '202406012315')
        self.assertEqual(minimal_timestr(T(20240601231523)), '20240601231523')
        self.assertEqual(minimal_timestr('202301031200'), '2023010312')
        self.assertEqual(minimal_timestr(None), None)

        self.assertListEqual(minimal_timestr(['202301', '2023021500']), ['20230101', '20230215'])
        arr_result = minimal_timestr(np.array(['202301', '2023021500']))
        self.assertIsInstance(arr_result, np.ndarray)
        self.assertListEqual(list(arr_result), ['20230101', '20230215'])

    def test_simple_timestr(self):
        self.assertEqual(simple_timestr(202305010000, freq='1Y'), '2023')
        self.assertEqual(simple_timestr(202305010000, freq='1M'), '202305')
        self.assertEqual(simple_timestr(202305010000, freq='1H'), '2023050100')
        self.assertEqual(simple_timestr(202305010000, freq='15m'), '202305010000')
        self.assertListEqual(simple_timestr([20230501, 20230502], freq='1D'), ['20230501', '20230502'])


if __name__ == '__main__':
    unittest.main()



__all__ = ['minimal_timestr', 'simple_timestr', 'TT']

from datetime import date, datetime
from .parsers import T
import numpy as np
from .common import return_as_type


def minimal_timestr(dt):
    if isinstance(dt, (list, tuple)):
        dts = dt
        return_type = 'list'
    elif isinstance(dt, np.ndarray):
        dts = dt
        return_type = 'np_arr'
    else:
        dts = [dt]
        return_type = 'scalar'

    res_list = []
    for dt in dts:
        if dt is None:
            res_list.append(None)
            continue

        if not isinstance(dt, date):
            dt = T(dt)
        timestr = dt.strftime('%Y%m%d%H%M%S')
        while timestr.endswith('00'):
            timestr = timestr[:-2]
        res_list.append(timestr)

    return return_as_type(res_list, return_type)


def simple_timestr(dt, freq):
    if isinstance(dt, (list, tuple)):
        dts = dt
        return_type = 'list'
    elif isinstance(dt, np.ndarray):
        dts = dt
        return_type = 'np_arr'
    else:
        dts = [dt]
        return_type = 'scalar'

    if freq.endswith(('T', 'min', 'm')):
        time_format = '%Y%m%d%H%M'
    elif freq.endswith(('H', 'h')):
        time_format = '%Y%m%d%H'
    elif freq.endswith(('D', 'd')):
        time_format = '%Y%m%d'
    elif freq.endswith(('M', 'MS')):
        time_format = '%Y%m'
    elif freq.endswith('Y'):
        time_format = '%Y'
    else:
        time_format = '%Y%m%d%H%M%S'

    res_list = []
    for dt in dts:
        if dt is None:
            res_list.append(None)
            continue

        if not isinstance(dt, date):
            dt = T(dt)
        res_list.append(dt.strftime(time_format))

    return return_as_type(res_list, return_type)


# 什么时候加的这个?
def TT(dt):
    if isinstance(dt, (list, tuple)):
        dts = dt
        return_type = 'list'
    elif isinstance(dt, np.ndarray):
        dts = dt
        return_type = 'np_arr'
    else:
        dts = [dt]
        return_type = 'scalar'

    result = []
    for item in dts:
        if item is None:
            result.append(None)
        elif isinstance(item, datetime):
            result.append(item.strftime("%Y-%m-%d %H:%M:%S"))
        elif isinstance(item, date):
            result.append(item.strftime("%Y-%m-%d"))
        else:
            result.append(T(item).strftime("%Y-%m-%d %H:%M:%S"))

    return return_as_type(result, return_type)



# -*- coding:utf-8 -*-

import six
import unittest
from hypothesis import given
import hypothesis.strategies as st
import hypothesis.extra.numpy as np_st

import numpy as np
from dt_utils.parsers import *


class TestParseDatetime(unittest.TestCase):
    @given(st.datetimes(min_value=datetime(1900, 1, 1)))
    def test_single_integer(self, raw_dt):
        for fmt in ('%Y%m%d', '%Y%m%d%H', '%Y%m%d%H%M', '%Y%m%d%H%M%S'):
            dt_num = int(raw_dt.strftime(fmt))
            parsed = T(dt_num)
            self.assertEqual(raw_dt.strftime(fmt), parsed.strftime(fmt), msg='Test Single Integer, fmt: {}, raw: {}, parsed: {}'.format(fmt, raw_dt, parsed))

    @given(st.datetimes(min_value=datetime(1900, 1, 1)))
    def test_single_str(self, raw_dt):
        for fmt in (
            '%Y%m%d', '%Y%m%d%H', '%Y%m%d%H%M', '%Y%m%d%H%M%S',
            '%Y-%m-%d', '%Y/%m/%d',
            '%Y-%m-%d %H',
            '%Y-%m-%d %H:%M',
            '%Y-%m-%dT%H:%M:%S',
            '%Y%j',
            '%Y%j%H'
        ):
            dt_str = raw_dt.strftime(fmt)
            parsed = T(dt_str)
            self.assertEqual(raw_dt.strftime(fmt), parsed.strftime(fmt), msg='Test Single String, fmt: {}, raw: {}, parsed: {}'.format(fmt, raw_dt, parsed))

    @given(st.lists(st.datetimes(min_value=datetime(1900, 1, 1))))
    def test_list(self, raw_dts):
        for fmt in ('%Y%m%d', '%Y%m%d%H', '%Y%m%d%H%M', '%Y%m%d%H%M%S'):
            raw_list = [dt.strftime(fmt) for dt in raw_dts]
            parsed = T(raw_list)
            self.assertEqual(type(parsed), list)
            parsed_strlist = [dt.strftime(fmt) for dt in parsed]
            self.assertListEqual(raw_list, parsed_strlist)

    @given(st.lists(st.datetimes(min_value=datetime(1900, 1, 1))))
    def test_array(self, raw_dts):
        for fmt in ('%Y%m%d', '%Y%m%d%H', '%Y%m%d%H%M', '%Y%m%d%H%M%S'):
            raw_list = [dt.strftime(fmt) for dt in raw_dts]
            raw_array = np.array(raw_list)
            parsed = T(raw_array)
            self.assertEqual(type(parsed), np.ndarray)
            parsed_strlist = [dt.strftime(fmt) for dt in parsed]
            self.assertListEqual(raw_list, parsed_strlist)

    def test_year_and_yearmonth(self):
        self.assertEqual(T('2024'), datetime(year=2024, month=1, day=1))
        self.assertEqual(T('202405'), datetime(year=2024, month=5, day=1))
        self.assertEqual(T('2024 05'), datetime(year=2024, month=5, day=1))
        self.assertEqual(T('2024-05'), datetime(year=2024, month=5, day=1))
        self.assertEqual(T('2024/05'), datetime(year=2024, month=5, day=1))
        self.assertEqual(T(2024), datetime(year=2024, month=1, day=1))
        self.assertEqual(T(202405), datetime(year=2024, month=5, day=1))


if __name__ == '__main__':
    unittest.main()


__all__ = ['get_latest_neat_time', 'get_nearest_neat_time', 'get_neat_beg_end']

import re
from datetime import datetime, timedelta
from .parsers import TD


def _get_n(s):
    return 1 if s == '' else int(s)


def get_latest_neat_time(time=None, freq='1H'):
    """获取最新的一个整齐时点
    :param time: 不提供时, 默认从当前时间算起
    :param freq:
    :return:
    """
    result = None
    if time is None:
        time = datetime.now()

    if m := re.match(r'^(\d*)(T|m|min)$', freq):
        n = _get_n(m.group(1))
        minute = (time.minute // n) * n
        result = time.replace(minute=minute, second=0, microsecond=0)
    elif m := re.match(r'^(\d*)([Hh])$', freq):
        n = _get_n(m.group(1))
        hour = (time.hour // n) * n
        result = time.replace(hour=hour, minute=0, second=0, microsecond=0)
    elif freq in ('d', 'D', '1d', '1D'):
        result = time.replace(hour=0, minute=0, second=0, microsecond=0)
    elif m := re.match(r'^(\d*)M$', freq):
        n = _get_n(m.group(1))
        month = ((time.month - 1) // n) * n + 1
        result = time.replace(month=month, day=1, hour=0, minute=0, second=0, microsecond=0)
    elif m := re.match(r'^(\d*)Y$', freq):
        n = _get_n(m.group(1))
        year = (time.year // n) * n
        result = time.replace(year=year, month=1, day=1, hour=0, minute=0, second=0, microsecond=0)

    if result is None:
        raise NotImplementedError(f"get_latest_neat_time of freq {freq} not implemented!")

    return result


def get_nearest_neat_time(time=None, freq='1H'):
    """获取距离最近的一个整齐时点
    :param time: 不提供时, 默认从当前时间算起
    :param freq:
    """
    if time is None:
        time = datetime.now()

    result = get_latest_neat_time(time=time, freq=freq)
    if m := re.match(r'^(\d*)M$', freq):
        n = _get_n(m.group(1))
        half_days = timedelta(days=int(31 * n / 2))
        if time - result >= half_days:
            result = result + TD(freq)
    elif m := re.match(r'^(\d*)Y$', freq):
        n = _get_n(m.group(1))
        half_days = timedelta(days=int(365 * n / 2))
        if time - result >= half_days:
            result = result + TD(freq)
    else:
        tdelta = TD(freq)
        if time - result > tdelta / 2:
            result = result + tdelta

    return result


def get_neat_beg_end(beg=None, end=None, freq='1H', offset=0, periods=1):
    """获取整齐的起止时间
    :param beg:
    :param end:
    :param freq:
    :param offset: 将结果平移几个周期
    :param periods: 当没有提供beg时, 给出时长为几个周期的时段
    """
    t_delta = TD(freq)
    neat_end = get_latest_neat_time(end, freq)
    neat_beg = neat_end - t_delta * periods if beg is None else get_latest_neat_time(beg, freq)
    if offset != 0:
        neat_beg += t_delta * offset
        neat_end += t_delta * offset
    return neat_beg, neat_end
